/* Freqt.java
   Copyright 2002, 
   Tatsuya Asai, Kenji Abe, Shinji Kawasoe, Hiroshi Sakamoto,
   Hiroki Arimura and Setsuo Arikawa,
   all rights reserved.

   This file contains an implementation of the algorithm "FREQT".

   This software may be used freely for any purpose. However, when distributed,
   the original source must be clearly stated, and, when the source code is
   distributed, the copiright notice must be retained and any alterations in
   the code must be clearly marked. No warranty is given regarding the quality
   of this software.*/

import java.io.*;
import java.lang.*;
import java.util.*;
import org.w3c.dom.*;

/**
 * FREQT 
 */
public class Freqt{

    static int stage    = 1;
    static int treesize = 0;    // Datatree Size
    static int gate     = 0;    // Minimum support
    static long totaltime = 0;	// Total Runtime (sec)
    static double minsup;	// Minimum support (%)
    static HashSet cutnode   = new HashSet();   // for Node-Skip
    static HashMap cutedge   = new HashMap();	// for Edge-Skip
    static HashMap concepttp = new HashMap();   // Concept Tree Path

    static final boolean ENABLE_PRUN_1NODE = false;
    static final boolean ENABLE_PRUN_2NODE = false;

    // Options and default settings
    static boolean ENABLE_DUPLICATE_DETECTION = true;
    static boolean ENABLE_ATTR_DOMTREE        = false;
    static boolean ENABLE_CONCEPT_NODE        = false;

    public static void main(String[] args){
        try{
	    if (args.length < 2)
		throw new IllegalArgumentException();

	    int i = 0;
            // Input Document
	    File pagefile = new File(args[i++]);
            // Minimum Support
	    double minsupper = Double.parseDouble(args[i++]);

	    // Setting for Options
	    // Default Options
	    File conceptfile = null;
	    int stagelimit = -1;

	    // Flag for option
	    boolean setDD, setAttr, setConcept, setLimit;
	    setDD = setAttr = setConcept = setLimit = false;

	    while (i < args.length){
		// Setting for Duplicate-Detection
		if (args[i].equals("-dd")){
		    if (setDD == true)
			throw new IllegalArgumentException();
		    i++;
		    ENABLE_DUPLICATE_DETECTION = true;
		    setDD = true;
		}
		else if (args[i].equals("-nodd")){
		    if (setDD == true)
			throw new IllegalArgumentException();
		    i++;
		    ENABLE_DUPLICATE_DETECTION = false;
		    setDD = true;
		}
		// Setting for Attribute
		else if (args[i].equals("-attr")){
		    if (setAttr == true)
			throw new IllegalArgumentException();
		    i++;
		    ENABLE_ATTR_DOMTREE = true;
		    setAttr = true;
		}
		else if (args[i].equals("-noattr")){
		    if (setAttr == true)
			throw new IllegalArgumentException();
		    i++;
		    ENABLE_ATTR_DOMTREE = false;
		    setAttr = true;
		}
		// Setting for Upper Concept
		else if (args[i].equals("-c")){
		    if (setConcept == true)
			throw new IllegalArgumentException();
		    i++;
		    ENABLE_CONCEPT_NODE = true;
		    conceptfile = new File(args[i++]);
		    setConcept = true;
		}
		// Setting for stage limit	
		else if (args[i].equals("-l")){
		    if (setLimit == true)
			throw new IllegalArgumentException();
		    i++;
		    stagelimit = Integer.parseInt(args[i++]);
		    setLimit = true;
		}
		// Other case
		else
		    throw new IllegalArgumentException();
	    }
	    
	    // Program Main Routine
	    execute(pagefile, minsupper, conceptfile, stagelimit);
        }

	catch(IllegalArgumentException e){
            System.err.println("Bad Parameters or Options.");
            System.err.println("USAGE: java Freqt [TARGET] [MINSUP] [OPTIONS]");
	    System.err.println(" TARGET is the input HTML/XML document.");
	    System.err.println(" MINSUP is the minimum-support specified by pacentage.");
	    System.err.println(" <OPTIONS>");
	    System.err.println("  -dd   (-nodd)      Enable (or Disable) Duplicate Detection");
	    System.err.println("  -attr (-noattr)    Enable (or Disable) Attribute Expand");
	    System.err.println("  -c <CONCEPT>       Set Concept Tree with <CONCEPT>");
	    System.err.println("  -l <MAXSTAGE>      Set Stage Limit with <MAXSTAGE>");
	    System.err.println(" No Options mean to add '-dd -noattr'.");
            System.exit(1);
        }

        catch(Exception e){
            System.out.println(e);
            System.exit(1);
        }
    }

    public static void execute(File pagefile, double minsupper, File conceptfile, int stagelimit)
	throws Exception{

	ArrayList plist;

	System.out.println("##Target Document= " + pagefile.getName());
	System.out.println("##File Size (byte)= " + pagefile.length());
	System.out.println("##Minimum Support (%)= " + minsupper);
	minsup = (minsupper * 0.01);

	Node root = initialStage(pagefile, conceptfile);
	// Case: No StageLimit
	if (stagelimit == -1){
	    plist = firststage(root);
	    stage++;

	    if (plist.size() > 0){
		plist = secondstage(plist);
		stage++;

		while (plist.size() > 0){
		    plist = laterstage(plist);
		    stage++;
		}
	    }
	}
	
	else{
	    if (stage <= stagelimit){
		plist = firststage(root);
		stage++;

		if (plist.size() > 0 && stage <= stagelimit){
		    plist = secondstage(plist);
		    stage++;
		    
		    while (plist.size() > 0 && stage <= stagelimit){
			plist = laterstage(plist);
			stage++;
		    }
		}
	    }
	}

	System.out.println("##Total CPU Time (sec)= " + (double)totaltime*0.001);
	System.out.println();
    }

    // Create DOM Tree Stage
    private static Node initialStage(File pagefile, File conceptfile) throws Exception{

        ArrayList plist = new ArrayList();

        Date time = new Date();
        long starttime = time.getTime();

	DomWrapper.initialize();
	Node root = DomWrapper.createDom(pagefile);
	Document doc = DomWrapper.getDocument();
	
	time = new Date();
	long parsetime = time.getTime() - starttime;

	System.out.println("##DOM Tree Parsing Stage (sec)= " + (double)parsetime*0.001);

	time = new Date();
	starttime = time.getTime();

	if (ENABLE_ATTR_DOMTREE)
	    DomWrapper.createAttrTree(root, doc);
	else
	    ;
	
	time = new Date();
	long parsetime2 = time.getTime() - starttime;

	System.out.println("##DOM Tree Attribute Expand Stage (sec)= " + (double)parsetime2*0.001);

	if (ENABLE_CONCEPT_NODE)
	    setConceptTreePath(conceptfile);
	else
	    ;

	return root;
    }
    
    private static ArrayList firststage(Node root){

	HashIndex tagList = new HashIndex();
	HashMultiIndex tagList2 = new HashMultiIndex();

        ArrayList plist = new ArrayList();

	Date time = new Date();
	long starttime = time.getTime();
	
	searchtree(root, tagList, tagList2, plist);
	setoccstage1(plist);

	plist = cutFirstStage(plist);
	
        time = new Date();
        long stagetime = time.getTime() - starttime;
        totaltime = totaltime + stagetime;

        ddisp(plist);

        System.out.println("##Stage " + stage + " CPU Time (sec)= " + (double)stagetime*0.001);
        System.out.println();
        return plist;
    }

    private static void searchtree(Node node, HashIndex tagList, HashMultiIndex tagList2, ArrayList plist){

        String nodename = node.getNodeName();
	treesize++;

     	if (ENABLE_CONCEPT_NODE == false){
	    int nid = tagList.get(nodename);
	    if (nid == plist.size())
		plist.add(new Pattern(nodename));
	    else
		;
	    Pattern pattern = (Pattern)plist.get(nid);
	    pattern.addNode(node);
	}

	else{
	    if (nodename.startsWith("[HTTP://") == true){
		String[] dns = UrlOperator.getGapDomains(nodename);
		int[] nids = tagList2.get(dns);
		for (int i = 0; i < nids.length; i++){
		    if (nids[i] == plist.size())
			plist.add(new Pattern(dns[i]));
		    else
			;
		    Pattern pattern = (Pattern)plist.get(nids[i]);
		    pattern.addNode(node);
		}
	    }
	    else{
		String[] conceptPath = getConceptTreePath(nodename);
		int[] nids = tagList2.get(nodename, conceptPath);
		for (int i = 0; i < nids.length; i++){
		    if (nids[i] == plist.size()){
			if (i == 0)
			    plist.add(new Pattern(nodename));
			else
			    plist.add(new Pattern(conceptPath[i-1]));
		    }
		    else
			;
		    Pattern pattern = (Pattern)plist.get(nids[i]);
		    pattern.addNode(node);
		}
	    }
	}

        Node nextnode = node.getFirstChild();
        while (nextnode != null){
            searchtree(nextnode, tagList, tagList2, plist);
            nextnode = nextnode.getNextSibling();
        }
    }

    private static ArrayList prunbyRoc(ArrayList plist){

	ArrayList newPlist = new ArrayList();

	for (int i = 0; i < plist.size(); i++){
	    Pattern pattern = (Pattern)plist.get(i);

	    if (pattern.getRMOcc() >= gate)
		newPlist.add(pattern);
	    else
		;
	}

	return newPlist;
    }

    private static void setoccstage1(ArrayList plist){

	for (int i = 0; i < plist.size(); i++){
	    Pattern pattern = (Pattern)plist.get(i);
	    pattern.setRootOcc(pattern.getRMOcc());
	}
    }

    private static int createocc(Pattern pattern){

	int rmb = pattern.getrmb();
	ArrayList roclist = pattern.getNodeList();
	// Variables for Duplicate-Detection
	Node lastocc = null;
	int occcount = 0;
	// Variable for Non Duplicate-Detection
	ArrayList occlist = new ArrayList();

	for (int i = 0; i < roclist.size(); i++){
	    Node target = (Node)roclist.get(i);
	    target = getAncestorNode(rmb, target);

	    if (ENABLE_DUPLICATE_DETECTION == true){
		if (target != lastocc){
		    occcount++;
		    lastocc = target;
		}
	    }

	    else{
		if (membernl(target, occlist) == false)
		    occlist.add(target);
	    }
	}

	if (ENABLE_DUPLICATE_DETECTION)
	    return occcount;
	else
	    return occlist.size();
    }

    private static Node getAncestorNode(int p, Node node){

	while (p > 0){
	    node = node.getParentNode();
	    p--;
	}
	return node;
    }

    private static void setocc(ArrayList plist){

	for (int i = 0; i < plist.size(); i++){
	    Pattern pattern = (Pattern)plist.get(i);
	    int rootocc = createocc(pattern);
	    pattern.setRootOcc(rootocc);
	}
    }

    private static ArrayList cutFirstStage(ArrayList plist){

	ArrayList newPlist = new ArrayList();

        System.out.println("##DataTree Size= " + treesize);
        gate = (int)Math.ceil(treesize * minsup);
        System.out.println("##Gate Size= " + gate);
        System.out.println();
	
        for (int i = 0; i < plist.size(); i++){
            Pattern pattern = (Pattern)plist.get(i);

            if (pattern.getRootOcc() >= gate)
		newPlist.add(pattern);

	    else{
		if (ENABLE_PRUN_1NODE){
		    Snode node = pattern.getPatNode();
		    String nodename = node.getNodeName();
		    addcutnode(nodename);
		}
	    }
        }

	return newPlist;
    }

    private static ArrayList secondstage(ArrayList plist){

        ArrayList newPlist = new ArrayList();

        Date time = new Date();
        long starttime = time.getTime();

        for (int i = 0; i < plist.size(); i++){
            Pattern pattern = (Pattern)plist.get(i);
	    Snode pnode = pattern.getPatNode();
            ArrayList nodeList = pattern.getNodeList();
	    createPatternStage2(pnode, nodeList, newPlist);
        }
       
	newPlist = prunbyRoc(newPlist);
	setocc(newPlist);
	newPlist = cutSecondStage(newPlist);

        time = new Date();
        long stagetime = time.getTime() - starttime;
        totaltime = totaltime + stagetime;

        ddisp(newPlist);

        System.out.println("##Stage " + stage + " CPU Time (sec)= " + (double)stagetime*0.001);
        System.out.println();

        return newPlist;
    }

    private static void createPatternStage2(Snode patnode, ArrayList nodeList, ArrayList plist){

	Snode newPnode;
	Pattern curPat;
	HashIndex hashList = new HashIndex(plist.size());
	HashMultiIndex hashList2 = new HashMultiIndex(plist.size());

	for (int i = 0; i < nodeList.size(); i++){
	    Node node = (Node)nodeList.get(i);
	    Node target = node.getFirstChild();

	    while (target != null){
		String nodename = target.getNodeName();

		if (ENABLE_CONCEPT_NODE == false){
		    if (ENABLE_PRUN_1NODE == false || askcutnode(nodename)){
			int nid = hashList.get(nodename);
			if (nid == plist.size()){
			    newPnode = createpattern(patnode, 0, nodename);
			    curPat = new Pattern(newPnode, 1);
			    plist.add(curPat);
			    curPat.addNode(target);
			}
			else{
			    curPat = (Pattern)plist.get(nid);
			    curPat.addNode(target);
			}
		    }
		}
		
		else{
		    if (nodename.startsWith("[HTTP://") == true){
			String[] dns = UrlOperator.getGapDomains(nodename);
			int[] nids = hashList2.get(dns);
			for (int j = 0; j < nids.length; j++){
			    if (nids[j] == plist.size()){
				newPnode = createpattern(patnode, 0, dns[j]);
				curPat = new Pattern(newPnode, 1);
				plist.add(curPat);
				curPat.addNode(target);
			    }
			    else{
				curPat = (Pattern)plist.get(nids[j]);
				curPat.addNode(target);
			    }
			}
		    }

		    else{	
			String[] conceptPath = getConceptTreePath(nodename);
			int[] nids = hashList2.get(nodename, conceptPath);

			for (int j = 0; j < nids.length; j++){
			    if (nids[j] == plist.size()){
				if (j == 0)
				    newPnode = createpattern(patnode, 0 ,nodename);
				else
				    newPnode = createpattern(patnode, 0, conceptPath[j-1]);
				curPat = new Pattern(newPnode,1);
				plist.add(curPat);
				curPat.addNode(target);
			    }
			    else{
				curPat = (Pattern)plist.get(nids[j]);
				curPat.addNode(target);
			    }
			}
		    }
		}
		target = target.getNextSibling();
	    }
	}
    }

    private static ArrayList cutSecondStage(ArrayList plist){

	ArrayList newPlist = new ArrayList();

        for (int i = 0; i < plist.size(); i++){
	    Pattern curPat = (Pattern)plist.get(i);

            if (curPat.getRootOcc() >= gate){
		newPlist.add(curPat);
	    }
            else{
		if (ENABLE_PRUN_2NODE){
		    Snode node = curPat.getPatNode();
		    String cname = node.getNodeName();
		    Snode pnode = node.getParent();
		    String pname = pnode.getNodeName();
		    addcutedge(pname,cname);
		}
	    }
        }

	return newPlist;
    }

    private static Snode createpattern(Snode ptree,int p,String nodename){

	Snode ntree = ptree.copypattern();
	ntree = ntree.expand(p,nodename);

	return ntree;
    }

    private static ArrayList laterstage(ArrayList plist){

        ArrayList newplist = new ArrayList();
        Date time = new Date();
        long starttime = time.getTime();

        for (int i = 0; i < plist.size(); i++){
            Pattern pattern = (Pattern)plist.get(i);
	    Snode patnode = pattern.getPatNode();
            ArrayList nodelist = pattern.getNodeList();
            int pmax = pattern.getrmb();

	    createpatstage3p0(patnode, nodelist, newplist, pmax);
	    for (int p = 1; p <= pmax; p++)
		createpatstage3(patnode, nodelist, newplist, pmax, p);
	}

	newplist = prunbyRoc(newplist);
	setocc(newplist);
        newplist = cutlaterstage(newplist);

        time = new Date();
        long stagetime = (time.getTime() - starttime);
        totaltime = (totaltime + stagetime);

        ddisp(newplist);

        System.out.println("##Stage " + stage + " CPU Time (sec)= " + (double)stagetime*0.001);
        System.out.println();

        return newplist;
    }

    private static void createpatstage3p0(Snode patnode, ArrayList nodelist, ArrayList plist, int rmb){

	int j, nid;
	Snode newpnode;
	HashIndex hashList = new HashIndex(plist.size());
	HashMultiIndex hashList2 = new HashMultiIndex(plist.size());

	for (int i = 0; i < nodelist.size(); i++){
	    Node node = (Node)nodelist.get(i);
	    Node target = node.getFirstChild();
	    String pname = node.getNodeName();
	    		
	    while (target != null){
		String nodename = target.getNodeName();

		if (ENABLE_CONCEPT_NODE == false){
		    if (ENABLE_PRUN_1NODE == false || askcutnode(nodename)){
			if (ENABLE_PRUN_2NODE == false || askcutedge(pname, nodename)){
			    nid = hashList.get(nodename);
			    
			    if (nid == plist.size()){
				newpnode = createpattern(patnode, 0, nodename);
				Pattern newpat = new Pattern(newpnode, rmb+1);
				plist.add(newpat);
				newpat.addNode(target);
			    }
			    else{
				Pattern curpat = (Pattern)plist.get(nid);
				curpat.addNode(target);
			    }
			}
		    }
		}

		else{
		    if (nodename.startsWith("[HTTP://") == true){
			String[] dns = UrlOperator.getGapDomains(nodename);
			int[] nids = hashList2.get(dns);
			for (j = 0; j < nids.length; j++){
			    if (nids[j] == plist.size()){
				newpnode = createpattern(patnode, 0, dns[j]);
				Pattern newpat = new Pattern(newpnode, rmb+1);
				plist.add(newpat);
				newpat.addNode(target);
			    }
			    else{
				Pattern curpat = (Pattern)plist.get(nids[j]);
				curpat.addNode(target);
			    }
			}
		    }

		    else{
			String[] conceptPath = getConceptTreePath(nodename);
			int[] nids = hashList2.get(nodename, conceptPath);

			for (j = 0; j < nids.length; j++){
			    if (nids[j] == plist.size()){
				if (j == 0)
				    newpnode = createpattern(patnode, 0, nodename);
				else
				    newpnode = createpattern(patnode, 0, conceptPath[j-1]);
				Pattern newpat = new Pattern(newpnode, rmb+1);
				plist.add(newpat);
				newpat.addNode(target);
			    }
			    else{
				Pattern curpat = (Pattern)plist.get(nids[j]);
				curpat.addNode(target);
			    }
			}
		    }
		}
		target = target.getNextSibling();
	    }
	}
    }
    
    private static void createpatstage3(Snode patnode, ArrayList nodelist, ArrayList plist, int rmb, int p){

	int j, nid;
	Snode newpnode;
	Node last = null;   // pth-parent
	HashIndex hashList = new HashIndex(plist.size());
	HashMultiIndex hashList2 = new HashMultiIndex(plist.size());

	for (int i = 0; i < nodelist.size(); i++){
	    Node node = (Node)nodelist.get(i);
	    Node target = getAncestorNode(p-1, node);
	    Node pnode = target.getParentNode();
	    String pname = pnode.getNodeName();
	    target = target.getNextSibling();

	    if (ENABLE_DUPLICATE_DETECTION){
		if (pnode != last){
		    while (target != null){
			String nodename = target.getNodeName();
			if (ENABLE_CONCEPT_NODE == false){
			    if (ENABLE_PRUN_1NODE == false || askcutnode(nodename)){
				if (ENABLE_PRUN_2NODE == false || askcutedge(pname, nodename)){
				    nid = hashList.get(nodename);

				    if (nid == plist.size()){
					newpnode = createpattern(patnode, p, nodename);
					Pattern newpat = new Pattern(newpnode, rmb - p + 1);
					plist.add(newpat);
					newpat.addNode(target);
				    }
				    else{
					Pattern curpat = (Pattern)plist.get(nid);
					curpat.addNode(target);
				    }
				}
			    }
			}

			else{
			    if (nodename.startsWith("[HTTP://") == true){
				String[] dns = UrlOperator.getGapDomains(nodename);
				int[] nids = hashList2.get(dns);
				for (j = 0; j < nids.length; j++){
				    if (nids[j] == plist.size()){
					newpnode = createpattern(patnode, p, dns[j]);
					Pattern newpat = new Pattern(newpnode, rmb - p + 1);
					plist.add(newpat);
					newpat.addNode(target);
				    }
				    else{
					Pattern curpat = (Pattern)plist.get(nids[j]);
					curpat.addNode(target);
				    }
				}
			    }

			    else{
				String[] conceptPath = getConceptTreePath(nodename);
				int[] nids = hashList2.get(nodename, conceptPath);
				for (j = 0; j < nids.length; j++){
				    if (nids[j] == plist.size()){
					if (j == 0)
					    newpnode = createpattern(patnode, p, nodename);
					else
					    newpnode = createpattern(patnode, p, conceptPath[j-1]);
					Pattern newpat = new Pattern(newpnode, rmb - p + 1);
					plist.add(newpat);
					newpat.addNode(target);
				    }
				    else{
					Pattern curpat = (Pattern)plist.get(nids[j]);
					curpat.addNode(target);
				    }
				}
			    }
			}
			target = target.getNextSibling();
		    }
		    last = pnode;
		}
	    }

	    else{
		while (target != null){
		    String nodename = target.getNodeName();

		    if (ENABLE_CONCEPT_NODE == false){
			if (ENABLE_PRUN_1NODE == false || askcutnode(nodename)){
			    if (ENABLE_PRUN_2NODE == false || askcutedge(pname, nodename)){
				nid = hashList.get(nodename);

				if (nid == plist.size()){
				    newpnode = createpattern(patnode, p, nodename);
				    Pattern newpat = new Pattern(newpnode, rmb - p + 1);
				    plist.add(newpat);
				    newpat.addNode(target);
				}
				else{
				    Pattern curpat = (Pattern)plist.get(nid);
				    ArrayList newnodelist = curpat.getNodeList();
				    if (membernl(target, newnodelist) == false)
					curpat.addNode(target);
				}
			    }
			}
		    }

		    else{
			if (nodename.startsWith("[HTTP://") == true){
			    String[] dns = UrlOperator.getGapDomains(nodename);
			    int[] nids = hashList2.get(dns);
			    for (j = 0; j < nids.length; j++){
				if (nids[j] == plist.size()){
				    newpnode = createpattern(patnode, p, dns[j]);
				    Pattern newpat = new Pattern(newpnode, rmb - p + 1);
				    plist.add(newpat);
				    newpat.addNode(target);
				}
				else{
				    Pattern curpat = (Pattern)plist.get(nids[j]);
				    ArrayList newnodelist = curpat.getNodeList();
				    if (membernl(target, newnodelist) == false)
					curpat.addNode(target);
				}
			    }
			}

			else{
			    String[] conceptPath = getConceptTreePath(nodename);
			    int[] nids = hashList2.get(nodename, conceptPath);
			    for (j = 0; j < nids.length; j++){
				if (nids[j] == plist.size()){
				    if (j == 0)
					newpnode = createpattern(patnode, p, nodename);
				    else
					newpnode = createpattern(patnode, p, conceptPath[j-1]);
				    Pattern newpat = new Pattern(newpnode, rmb - p + 1);
				    plist.add(newpat);
				    newpat.addNode(target);
				}
				else{
				    Pattern curpat = (Pattern)plist.get(nids[j]);
				    ArrayList newnodelist = curpat.getNodeList();
				    if (membernl(target, newnodelist) == false)
					curpat.addNode(target);
				}
			    }
			}
		    }
		    target = target.getNextSibling();
		}
	    }
	}
    }
    
    private static ArrayList cutlaterstage(ArrayList plist){

	ArrayList newPlist = new ArrayList();
	
        for (int i = 0; i < plist.size(); i++){
	    Pattern pattern = (Pattern)plist.get(i);

            if (pattern.getRootOcc() >= gate){
		newPlist.add(pattern);
	    }
	    else
		;
	}

	return newPlist;
    }
	
    private static void ddisp(ArrayList plist){

	System.out.println("=== Stage " + stage + " ===");
	
	for (int i = 0; i < plist.size(); i++){
	    Pattern pattern = (Pattern)plist.get(i);
	    Snode patnode = pattern.getPatNode();
	    
	    System.out.print("Tree: ");
	    System.out.print(patnode.printTagStyle());
	    System.out.println();
	    
	    ArrayList nodelist = pattern.getNodeList();
	    System.out.print("No: " + i + "\tSize: " + stage);
	    System.out.println("\tOcc(Root/RM): " + pattern.getRootOcc() + "/ " + pattern.getRMOcc());
	}
	System.out.println();
	System.out.println("##Stage " + stage + " Patterns Total= " + plist.size());
	
    }

    private static void addcutnode(String nodename){

        cutnode.add(new String(nodename));
    }

    private static boolean askcutnode(String nodename){

        if (cutnode.contains(nodename))
            return false;
        else
            return true;
    }

    private static void addcutedge(String pname,String cname){

        if (cutedge.containsKey(pname) == false)
	    cutedge.put(new String(pname),new HashSet());

	HashSet cutchild = (HashSet)cutedge.get(pname);
        cutchild.add(new String(cname));
    }

    private static boolean askcutedge(String pname,String cname){

        if (cutedge.containsKey(pname)){
	    HashSet cutchild = (HashSet)cutedge.get(pname);
            if (cutchild.contains(cname))
                return false;
            else
                return true;
        }
        else
            return true;
    }

    private static boolean membernl(Node target, ArrayList nodelist){

        int i;
        Node current;

        for (i = (nodelist.size()-1); i >= 0; i--){
            current = (Node)nodelist.get(i);
            if (current == target)
                return true;
        }
        return false;
    }

    private static boolean membernn(String nodename,ArrayList al){
	
        for (int i = 0; i < al.size(); i++){
            String name = (String)al.get(i);
            if (name.equals(nodename))
                return true;
        }

        return false;
    }

    public static void setConceptTreePath(File file) throws Exception{

	int i;
	String s,key;
	String[] value;
	BufferedReader fin = new BufferedReader(new FileReader(file));
	while ((s=fin.readLine()) != null){
	    i = 0;
	    StringTokenizer st = new StringTokenizer(s," \t");
	    if (st.countTokens() > 1){
		key = st.nextToken();
		value = new String[st.countTokens()];
		while (st.countTokens() > 0)
		    value[i++] = st.nextToken();
		concepttp.put(key,value);
	    }
	}
	fin.close();
    }
	
    public static String[] getConceptTreePath(String tagname){

	if (concepttp.containsKey(tagname) == true)
	    return (String[])concepttp.get(tagname);
	else
	    return new String[]{};
    }
}





